DEBUG = 10
INFO = 20
SUCCESS = 25
WARNING = 30
ERROR = 40

DEFAULT_TAGS = {
    DEBUG: 'debug',
    INFO: 'info',
    SUCCESS: 'success',
    WARNING: 'warning',
    ERROR: 'error',
}

DEFAULT_LEVELS = {
    'DEBUG': DEBUG,
    'INFO': INFO,
    'SUCCESS': SUCCESS,
    'WARNING': WARNING,
    'ERROR': ERROR,
}



from django.db.models.expressions import F, OrderBy


class OrderableAggMixin:

    def __init__(self, expression, ordering=(), **extra):
        if not isinstance(ordering, (list, tuple)):
            ordering = [ordering]
        ordering = ordering or []
        # Transform minus sign prefixed strings into an OrderBy() expression.
        ordering = (
            (OrderBy(F(o[1:]), descending=True) if isinstance(o, str) and o[0] == '-' else o)
            for o in ordering
        )
        super().__init__(expression, **extra)
        self.ordering = self._parse_expressions(*ordering)

    def resolve_expression(self, *args, **kwargs):
        self.ordering = [expr.resolve_expression(*args, **kwargs) for expr in self.ordering]
        return super().resolve_expression(*args, **kwargs)

    def as_sql(self, compiler, connection):
        if self.ordering:
            ordering_params = []
            ordering_expr_sql = []
            for expr in self.ordering:
                expr_sql, expr_params = expr.as_sql(compiler, connection)
                ordering_expr_sql.append(expr_sql)
                ordering_params.extend(expr_params)
            sql, sql_params = super().as_sql(compiler, connection, ordering=(
                'ORDER BY ' + ', '.join(ordering_expr_sql)
            ))
            return sql, sql_params + ordering_params
        return super().as_sql(compiler, connection, ordering='')

    def set_source_expressions(self, exprs):
        # Extract the ordering expressions because ORDER BY clause is handled
        # in a custom way.
        self.ordering = exprs[self._get_ordering_expressions_index():]
        return super().set_source_expressions(exprs[:self._get_ordering_expressions_index()])

    def get_source_expressions(self):
        return self.source_expressions + self.ordering

    def get_source_fields(self):
        # Filter out fields contributed by the ordering expressions as
        # these should not be used to determine which the return type of the
        # expression.
        return [
            e._output_field_or_none
            for e in self.get_source_expressions()[:self._get_ordering_expressions_index()]
        ]

    def _get_ordering_expressions_index(self):
        """Return the index at which the ordering expressions start."""
        source_expressions = self.get_source_expressions()
        return len(source_expressions) - len(self.ordering)



from django.apps import AppConfig
from django.contrib.contenttypes.checks import (
    check_generic_foreign_keys, check_model_name_lengths,
)
from django.core import checks
from django.db.models.signals import post_migrate, pre_migrate
from django.utils.translation import gettext_lazy as _

from .management import (
    create_contenttypes, inject_rename_contenttypes_operations,
)


class ContentTypesConfig(AppConfig):
    name = 'django.contrib.contenttypes'
    verbose_name = _("Content Types")

    def ready(self):
        pre_migrate.connect(inject_rename_contenttypes_operations, sender=self)
        post_migrate.connect(create_contenttypes)
        checks.register(check_generic_foreign_keys, checks.Tags.models)
        checks.register(check_model_name_lengths, checks.Tags.models)



print("import: 'django'")
import django

print("import: 'django.apps'")
import django.apps

print("import: 'django.conf'")
import django.conf

print("import: 'django.conf.app_template.migrations'")
import django.conf.app_template.migrations

print("import: 'django.conf.locale'")
import django.conf.locale

print("import: 'django.conf.locale.ar'")
import django.conf.locale.ar

print("import: 'django.conf.locale.az'")
import django.conf.locale.az

print("import: 'django.conf.locale.bg'")
import django.conf.locale.bg

print("import: 'django.conf.locale.bn'")
import django.conf.locale.bn

print("import: 'django.conf.locale.bs'")
import django.conf.locale.bs

print("import: 'django.conf.locale.ca'")
import django.conf.locale.ca

print("import: 'django.conf.locale.cs'")
import django.conf.locale.cs

print("import: 'django.conf.locale.cy'")
import django.conf.locale.cy

print("import: 'django.conf.locale.da'")
import django.conf.locale.da

print("import: 'django.conf.locale.de'")
import django.conf.locale.de

print("import: 'django.conf.locale.de_CH'")
import django.conf.locale.de_CH

print("import: 'django.conf.locale.el'")
import django.conf.locale.el

print("import: 'django.conf.locale.en'")
import django.conf.locale.en

print("import: 'django.conf.locale.en_AU'")
import django.conf.locale.en_AU

print("import: 'django.conf.locale.en_GB'")
import django.conf.locale.en_GB

print("import: 'django.conf.locale.eo'")
import django.conf.locale.eo

print("import: 'django.conf.locale.es'")
import django.conf.locale.es

print("import: 'django.conf.locale.es_AR'")
import django.conf.locale.es_AR

print("import: 'django.conf.locale.es_CO'")
import django.conf.locale.es_CO

print("import: 'django.conf.locale.es_MX'")
import django.conf.locale.es_MX

print("import: 'django.conf.locale.es_NI'")
import django.conf.locale.es_NI

print("import: 'django.conf.locale.es_PR'")
import django.conf.locale.es_PR

print("import: 'django.conf.locale.et'")
import django.conf.locale.et

print("import: 'django.conf.locale.eu'")
import django.conf.locale.eu

print("import: 'django.conf.locale.fa'")
import django.conf.locale.fa

print("import: 'django.conf.locale.fi'")
import django.conf.locale.fi

print("import: 'django.conf.locale.fr'")
import django.conf.locale.fr

print("import: 'django.conf.locale.fy'")
import django.conf.locale.fy

print("import: 'django.conf.locale.ga'")
import django.conf.locale.ga

print("import: 'django.conf.locale.gd'")
import django.conf.locale.gd

print("import: 'django.conf.locale.gl'")
import django.conf.locale.gl

print("import: 'django.conf.locale.he'")
import django.conf.locale.he

print("import: 'django.conf.locale.hi'")
import django.conf.locale.hi

print("import: 'django.conf.locale.hr'")
import django.conf.locale.hr

print("import: 'django.conf.locale.hu'")
import django.conf.locale.hu

print("import: 'django.conf.locale.id'")
import django.conf.locale.id

print("import: 'django.conf.locale.it'")
import django.conf.locale.it

print("import: 'django.conf.locale.ja'")
import django.conf.locale.ja

print("import: 'django.conf.locale.ka'")
import django.conf.locale.ka

print("import: 'django.conf.locale.km'")
import django.conf.locale.km

print("import: 'django.conf.locale.kn'")
import django.conf.locale.kn

print("import: 'django.conf.locale.ko'")
import django.conf.locale.ko

print("import: 'django.conf.locale.lt'")
import django.conf.locale.lt

print("import: 'django.conf.locale.lv'")
import django.conf.locale.lv

print("import: 'django.conf.locale.mk'")
import django.conf.locale.mk

print("import: 'django.conf.locale.ml'")
import django.conf.locale.ml

print("import: 'django.conf.locale.mn'")
import django.conf.locale.mn

print("import: 'django.conf.locale.nb'")
import django.conf.locale.nb

print("import: 'django.conf.locale.nl'")
import django.conf.locale.nl

print("import: 'django.conf.locale.nn'")
import django.conf.locale.nn

print("import: 'django.conf.locale.pl'")
import django.conf.locale.pl

print("import: 'django.conf.locale.pt'")
import django.conf.locale.pt

print("import: 'django.conf.locale.pt_BR'")
import django.conf.locale.pt_BR

print("import: 'django.conf.locale.ro'")
import django.conf.locale.ro

print("import: 'django.conf.locale.ru'")
import django.conf.locale.ru

print("import: 'django.conf.locale.sk'")
import django.conf.locale.sk

print("import: 'django.conf.locale.sl'")
import django.conf.locale.sl

print("import: 'django.conf.locale.sq'")
import django.conf.locale.sq

print("import: 'django.conf.locale.sr'")
import django.conf.locale.sr

print("import: 'django.conf.locale.sr_Latn'")
import django.conf.locale.sr_Latn

print("import: 'django.conf.locale.sv'")
import django.conf.locale.sv

print("import: 'django.conf.locale.ta'")
import django.conf.locale.ta

print("import: 'django.conf.locale.te'")
import django.conf.locale.te

print("import: 'django.conf.locale.th'")
import django.conf.locale.th

print("import: 'django.conf.locale.tr'")
import django.conf.locale.tr

print("import: 'django.conf.locale.uk'")
import django.conf.locale.uk

print("import: 'django.conf.locale.vi'")
import django.conf.locale.vi

print("import: 'django.conf.locale.zh_Hans'")
import django.conf.locale.zh_Hans

print("import: 'django.conf.locale.zh_Hant'")
import django.conf.locale.zh_Hant

print("import: 'django.conf.urls'")
import django.conf.urls

print("import: 'django.contrib'")
import django.contrib

print("import: 'django.contrib.admin'")
import django.contrib.admin

print("import: 'django.contrib.admin.migrations'")
import django.contrib.admin.migrations

print("import: 'django.contrib.admin.templatetags'")
import django.contrib.admin.templatetags

print("import: 'django.contrib.admin.views'")
import django.contrib.admin.views

print("import: 'django.contrib.admindocs'")
import django.contrib.admindocs

print("import: 'django.contrib.auth'")
import django.contrib.auth

print("import: 'django.contrib.auth.handlers'")
import django.contrib.auth.handlers

print("import: 'django.contrib.auth.management'")
import django.contrib.auth.management

print("import: 'django.contrib.auth.management.commands'")
import django.contrib.auth.management.commands

print("import: 'django.contrib.auth.migrations'")
import django.contrib.auth.migrations

print("import: 'django.contrib.contenttypes'")
import django.contrib.contenttypes

print("import: 'django.contrib.contenttypes.management'")
import django.contrib.contenttypes.management

print("import: 'django.contrib.contenttypes.management.commands'")
import django.contrib.contenttypes.management.commands

print("import: 'django.contrib.contenttypes.migrations'")
import django.contrib.contenttypes.migrations

print("import: 'django.contrib.flatpages'")
import django.contrib.flatpages

print("import: 'django.contrib.flatpages.migrations'")
import django.contrib.flatpages.migrations

print("import: 'django.contrib.flatpages.templatetags'")
import django.contrib.flatpages.templatetags

print("import: 'django.contrib.gis'")
import django.contrib.gis

print("import: 'django.contrib.gis.db'")
import django.contrib.gis.db

print("import: 'django.contrib.gis.db.backends'")
import django.contrib.gis.db.backends

print("import: 'django.contrib.gis.db.backends.base'")
import django.contrib.gis.db.backends.base

print("import: 'django.contrib.gis.db.backends.mysql'")
import django.contrib.gis.db.backends.mysql

print("import: 'django.contrib.gis.db.backends.oracle'")
import django.contrib.gis.db.backends.oracle

print("import: 'django.contrib.gis.db.backends.postgis'")
import django.contrib.gis.db.backends.postgis

print("import: 'django.contrib.gis.db.backends.spatialite'")
import django.contrib.gis.db.backends.spatialite

print("import: 'django.contrib.gis.geometry'")
import django.contrib.gis.geometry

print("import: 'django.contrib.gis.management'")
import django.contrib.gis.management

print("import: 'django.contrib.gis.management.commands'")
import django.contrib.gis.management.commands

print("import: 'django.contrib.gis.serializers'")
import django.contrib.gis.serializers

print("import: 'django.contrib.humanize'")
import django.contrib.humanize

print("import: 'django.contrib.humanize.templatetags'")
import django.contrib.humanize.templatetags

print("import: 'django.contrib.messages'")
import django.contrib.messages

print("import: 'django.contrib.messages.storage'")
import django.contrib.messages.storage

print("import: 'django.contrib.postgres'")
import django.contrib.postgres

print("import: 'django.contrib.postgres.forms'")
import django.contrib.postgres.forms

print("import: 'django.contrib.redirects'")
import django.contrib.redirects

print("import: 'django.contrib.redirects.migrations'")
import django.contrib.redirects.migrations

print("import: 'django.contrib.sessions'")
import django.contrib.sessions

print("import: 'django.contrib.sessions.backends'")
import django.contrib.sessions.backends

print("import: 'django.contrib.sessions.management'")
import django.contrib.sessions.management

print("import: 'django.contrib.sessions.management.commands'")
import django.contrib.sessions.management.commands

print("import: 'django.contrib.sessions.migrations'")
import django.contrib.sessions.migrations

print("import: 'django.contrib.sitemaps'")
import django.contrib.sitemaps

print("import: 'django.contrib.sitemaps.management'")
import django.contrib.sitemaps.management

print("import: 'django.contrib.sitemaps.management.commands'")
import django.contrib.sitemaps.management.commands

print("import: 'django.contrib.sites'")
import django.contrib.sites

print("import: 'django.contrib.sites.migrations'")
import django.contrib.sites.migrations

print("import: 'django.contrib.staticfiles'")
import django.contrib.staticfiles

print("import: 'django.contrib.staticfiles.management'")
import django.contrib.staticfiles.management

print("import: 'django.contrib.staticfiles.management.commands'")
import django.contrib.staticfiles.management.commands

print("import: 'django.contrib.staticfiles.templatetags'")
import django.contrib.staticfiles.templatetags

print("import: 'django.contrib.syndication'")
import django.contrib.syndication

print("import: 'django.core'")
import django.core

print("import: 'django.core.cache'")
import django.core.cache

print("import: 'django.core.cache.backends'")
import django.core.cache.backends

print("import: 'django.core.checks'")
import django.core.checks

print("import: 'django.core.checks.compatibility'")
import django.core.checks.compatibility

print("import: 'django.core.checks.security'")
import django.core.checks.security

print("import: 'django.core.files'")
import django.core.files

print("import: 'django.core.handlers'")
import django.core.handlers

print("import: 'django.core.mail'")
import django.core.mail

print("import: 'django.core.mail.backends'")
import django.core.mail.backends

print("import: 'django.core.management'")
import django.core.management

print("import: 'django.core.management.commands'")
import django.core.management.commands

print("import: 'django.core.serializers'")
import django.core.serializers

print("import: 'django.core.servers'")
import django.core.servers

print("import: 'django.db'")
import django.db

print("import: 'django.db.backends'")
import django.db.backends

print("import: 'django.db.backends.base'")
import django.db.backends.base

print("import: 'django.db.backends.dummy'")
import django.db.backends.dummy

print("import: 'django.db.backends.mysql'")
import django.db.backends.mysql

print("import: 'django.db.backends.oracle'")
import django.db.backends.oracle

print("import: 'django.db.backends.postgresql'")
import django.db.backends.postgresql

print("import: 'django.db.backends.postgresql_psycopg2'")
import django.db.backends.postgresql_psycopg2

print("import: 'django.db.backends.sqlite3'")
import django.db.backends.sqlite3

print("import: 'django.db.migrations'")
import django.db.migrations

print("import: 'django.db.migrations.operations'")
import django.db.migrations.operations

print("import: 'django.db.models'")
import django.db.models

print("import: 'django.db.models.fields'")
import django.db.models.fields

print("import: 'django.db.models.functions'")
import django.db.models.functions

print("import: 'django.db.models.sql'")
import django.db.models.sql

print("import: 'django.dispatch'")
import django.dispatch

print("import: 'django.forms'")
import django.forms

print("import: 'django.http'")
import django.http

print("import: 'django.middleware'")
import django.middleware

print("import: 'django.template'")
import django.template

print("import: 'django.template.backends'")
import django.template.backends

print("import: 'django.template.loaders'")
import django.template.loaders

print("import: 'django.templatetags'")
import django.templatetags

print("import: 'django.test'")
import django.test

print("import: 'django.urls'")
import django.urls

print("import: 'django.utils'")
import django.utils

print("import: 'django.utils.translation'")
import django.utils.translation

print("import: 'django.views'")
import django.views

print("import: 'django.views.decorators'")
import django.views.decorators

print("import: 'django.views.generic'")
import django.views.generic




"""
GDAL - Constant definitions
"""
from ctypes import (
    c_double, c_float, c_int16, c_int32, c_ubyte, c_uint16, c_uint32,
)

# See https://www.gdal.org/gdal_8h.html#a22e22ce0a55036a96f652765793fb7a4
GDAL_PIXEL_TYPES = {
    0: 'GDT_Unknown',  # Unknown or unspecified type
    1: 'GDT_Byte',  # Eight bit unsigned integer
    2: 'GDT_UInt16',  # Sixteen bit unsigned integer
    3: 'GDT_Int16',  # Sixteen bit signed integer
    4: 'GDT_UInt32',  # Thirty-two bit unsigned integer
    5: 'GDT_Int32',  # Thirty-two bit signed integer
    6: 'GDT_Float32',  # Thirty-two bit floating point
    7: 'GDT_Float64',  # Sixty-four bit floating point
    8: 'GDT_CInt16',  # Complex Int16
    9: 'GDT_CInt32',  # Complex Int32
    10: 'GDT_CFloat32',  # Complex Float32
    11: 'GDT_CFloat64',  # Complex Float64
}

# A list of gdal datatypes that are integers.
GDAL_INTEGER_TYPES = [1, 2, 3, 4, 5]

# Lookup values to convert GDAL pixel type indices into ctypes objects.
# The GDAL band-io works with ctypes arrays to hold data to be written
# or to hold the space for data to be read into. The lookup below helps
# selecting the right ctypes object for a given gdal pixel type.
GDAL_TO_CTYPES = [
    None, c_ubyte, c_uint16, c_int16, c_uint32, c_int32,
    c_float, c_double, None, None, None, None
]

# List of resampling algorithms that can be used to warp a GDALRaster.
GDAL_RESAMPLE_ALGORITHMS = {
    'NearestNeighbour': 0,
    'Bilinear': 1,
    'Cubic': 2,
    'CubicSpline': 3,
    'Lanczos': 4,
    'Average': 5,
    'Mode': 6,
}

# See https://www.gdal.org/gdal_8h.html#ace76452d94514561fffa8ea1d2a5968c
GDAL_COLOR_TYPES = {
    0: 'GCI_Undefined',  # Undefined, default value, i.e. not known
    1: 'GCI_GrayIndex',  # Greyscale
    2: 'GCI_PaletteIndex',  # Paletted
    3: 'GCI_RedBand',  # Red band of RGBA image
    4: 'GCI_GreenBand',  # Green band of RGBA image
    5: 'GCI_BlueBand',  # Blue band of RGBA image
    6: 'GCI_AlphaBand',  # Alpha (0=transparent, 255=opaque)
    7: 'GCI_HueBand',  # Hue band of HLS image
    8: 'GCI_SaturationBand',  # Saturation band of HLS image
    9: 'GCI_LightnessBand',  # Lightness band of HLS image
    10: 'GCI_CyanBand',  # Cyan band of CMYK image
    11: 'GCI_MagentaBand',  # Magenta band of CMYK image
    12: 'GCI_YellowBand',  # Yellow band of CMYK image
    13: 'GCI_BlackBand',  # Black band of CMLY image
    14: 'GCI_YCbCr_YBand',  # Y Luminance
    15: 'GCI_YCbCr_CbBand',  # Cb Chroma
    16: 'GCI_YCbCr_CrBand',  # Cr Chroma, also GCI_Max
}

# Fixed base path for buffer-based GDAL in-memory files.
VSI_FILESYSTEM_BASE_PATH = '/vsimem/'

# Should the memory file system take ownership of the buffer, freeing it when
# the file is deleted? (No, GDALRaster.__del__() will delete the buffer.)
VSI_TAKE_BUFFER_OWNERSHIP = False

# Should a VSI file be removed when retrieving its buffer?
VSI_DELETE_BUFFER_ON_READ = False



import logging
import types

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, MiddlewareNotUsed
from django.core.signals import request_finished
from django.db import connections, transaction
from django.urls import get_resolver, set_urlconf
from django.utils.log import log_response
from django.utils.module_loading import import_string

from .exception import convert_exception_to_response

logger = logging.getLogger('django.request')


class BaseHandler:
    _view_middleware = None
    _template_response_middleware = None
    _exception_middleware = None
    _middleware_chain = None

    def load_middleware(self):
        """
        Populate middleware lists from settings.MIDDLEWARE.

        Must be called after the environment is fixed (see __call__ in subclasses).
        """
        self._view_middleware = []
        self._template_response_middleware = []
        self._exception_middleware = []

        handler = convert_exception_to_response(self._get_response)
        for middleware_path in reversed(settings.MIDDLEWARE):
            middleware = import_string(middleware_path)
            try:
                mw_instance = middleware(handler)
            except MiddlewareNotUsed as exc:
                if settings.DEBUG:
                    if str(exc):
                        logger.debug('MiddlewareNotUsed(%r): %s', middleware_path, exc)
                    else:
                        logger.debug('MiddlewareNotUsed: %r', middleware_path)
                continue

            if mw_instance is None:
                raise ImproperlyConfigured(
                    'Middleware factory %s returned None.' % middleware_path
                )

            if hasattr(mw_instance, 'process_view'):
                self._view_middleware.insert(0, mw_instance.process_view)
            if hasattr(mw_instance, 'process_template_response'):
                self._template_response_middleware.append(mw_instance.process_template_response)
            if hasattr(mw_instance, 'process_exception'):
                self._exception_middleware.append(mw_instance.process_exception)

            handler = convert_exception_to_response(mw_instance)

        # We only assign to this when initialization is complete as it is used
        # as a flag for initialization being complete.
        self._middleware_chain = handler

    def make_view_atomic(self, view):
        non_atomic_requests = getattr(view, '_non_atomic_requests', set())
        for db in connections.all():
            if db.settings_dict['ATOMIC_REQUESTS'] and db.alias not in non_atomic_requests:
                view = transaction.atomic(using=db.alias)(view)
        return view

    def get_response(self, request):
        """Return an HttpResponse object for the given HttpRequest."""
        # Setup default url resolver for this thread
        set_urlconf(settings.ROOT_URLCONF)
        response = self._middleware_chain(request)
        response._closable_objects.append(request)
        if response.status_code >= 400:
            log_response(
                '%s: %s', response.reason_phrase, request.path,
                response=response,
                request=request,
            )
        return response

    def _get_response(self, request):
        """
        Resolve and call the view, then apply view, exception, and
        template_response middleware. This method is everything that happens
        inside the request/response middleware.
        """
        response = None

        if hasattr(request, 'urlconf'):
            urlconf = request.urlconf
            set_urlconf(urlconf)
            resolver = get_resolver(urlconf)
        else:
            resolver = get_resolver()

        resolver_match = resolver.resolve(request.path_info)
        callback, callback_args, callback_kwargs = resolver_match
        request.resolver_match = resolver_match

        # Apply view middleware
        for middleware_method in self._view_middleware:
            response = middleware_method(request, callback, callback_args, callback_kwargs)
            if response:
                break

        if response is None:
            wrapped_callback = self.make_view_atomic(callback)
            try:
                response = wrapped_callback(request, *callback_args, **callback_kwargs)
            except Exception as e:
                response = self.process_exception_by_middleware(e, request)

        # Complain if the view returned None (a common error).
        if response is None:
            if isinstance(callback, types.FunctionType):    # FBV
                view_name = callback.__name__
            else:                                           # CBV
                view_name = callback.__class__.__name__ + '.__call__'

            raise ValueError(
                "The view %s.%s didn't return an HttpResponse object. It "
                "returned None instead." % (callback.__module__, view_name)
            )

        # If the response supports deferred rendering, apply template
        # response middleware and then render the response
        elif hasattr(response, 'render') and callable(response.render):
            for middleware_method in self._template_response_middleware:
                response = middleware_method(request, response)
                # Complain if the template response middleware returned None (a common error).
                if response is None:
                    raise ValueError(
                        "%s.process_template_response didn't return an "
                        "HttpResponse object. It returned None instead."
                        % (middleware_method.__self__.__class__.__name__)
                    )

            try:
                response = response.render()
            except Exception as e:
                response = self.process_exception_by_middleware(e, request)

        return response

    def process_exception_by_middleware(self, exception, request):
        """
        Pass the exception to the exception middleware. If no middleware
        return a response for this exception, raise it.
        """
        for middleware_method in self._exception_middleware:
            response = middleware_method(request, exception)
            if response:
                return response
        raise


def reset_urlconf(sender, **kwargs):
    """Reset the URLconf after each request is finished."""
    set_urlconf(None)


request_finished.connect(reset_urlconf)



import copy
import datetime
import re

from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.utils import DatabaseError


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):

    sql_create_column = "ALTER TABLE %(table)s ADD %(column)s %(definition)s"
    sql_alter_column_type = "MODIFY %(column)s %(type)s"
    sql_alter_column_null = "MODIFY %(column)s NULL"
    sql_alter_column_not_null = "MODIFY %(column)s NOT NULL"
    sql_alter_column_default = "MODIFY %(column)s DEFAULT %(default)s"
    sql_alter_column_no_default = "MODIFY %(column)s DEFAULT NULL"
    sql_delete_column = "ALTER TABLE %(table)s DROP COLUMN %(column)s"
    sql_delete_table = "DROP TABLE %(table)s CASCADE CONSTRAINTS"
    sql_create_index = "CREATE INDEX %(name)s ON %(table)s (%(columns)s)%(extra)s"

    def quote_value(self, value):
        if isinstance(value, (datetime.date, datetime.time, datetime.datetime)):
            return "'%s'" % value
        elif isinstance(value, str):
            return "'%s'" % value.replace("\'", "\'\'").replace('%', '%%')
        elif isinstance(value, (bytes, bytearray, memoryview)):
            return "'%s'" % value.hex()
        elif isinstance(value, bool):
            return "1" if value else "0"
        else:
            return str(value)

    def remove_field(self, model, field):
        # If the column is an identity column, drop the identity before
        # removing the field.
        if self._is_identity_column(model._meta.db_table, field.column):
            self._drop_identity(model._meta.db_table, field.column)
        super().remove_field(model, field)

    def delete_model(self, model):
        # Run superclass action
        super().delete_model(model)
        # Clean up manually created sequence.
        self.execute("""
            DECLARE
                i INTEGER;
            BEGIN
                SELECT COUNT(1) INTO i FROM USER_SEQUENCES
                    WHERE SEQUENCE_NAME = '%(sq_name)s';
                IF i = 1 THEN
                    EXECUTE IMMEDIATE 'DROP SEQUENCE "%(sq_name)s"';
                END IF;
            END;
        /""" % {'sq_name': self.connection.ops._get_no_autofield_sequence_name(model._meta.db_table)})

    def alter_field(self, model, old_field, new_field, strict=False):
        try:
            super().alter_field(model, old_field, new_field, strict)
        except DatabaseError as e:
            description = str(e)
            # If we're changing type to an unsupported type we need a
            # SQLite-ish workaround
            if 'ORA-22858' in description or 'ORA-22859' in description:
                self._alter_field_type_workaround(model, old_field, new_field)
            # If an identity column is changing to a non-numeric type, drop the
            # identity first.
            elif 'ORA-30675' in description:
                self._drop_identity(model._meta.db_table, old_field.column)
                self.alter_field(model, old_field, new_field, strict)
            # If a primary key column is changing to an identity column, drop
            # the primary key first.
            elif 'ORA-30673' in description and old_field.primary_key:
                self._delete_primary_key(model, strict=True)
                self._alter_field_type_workaround(model, old_field, new_field)
            else:
                raise

    def _alter_field_type_workaround(self, model, old_field, new_field):
        """
        Oracle refuses to change from some type to other type.
        What we need to do instead is:
        - Add a nullable version of the desired field with a temporary name. If
          the new column is an auto field, then the temporary column can't be
          nullable.
        - Update the table to transfer values from old to new
        - Drop old column
        - Rename the new column and possibly drop the nullable property
        """
        # Make a new field that's like the new one but with a temporary
        # column name.
        new_temp_field = copy.deepcopy(new_field)
        new_temp_field.null = (new_field.get_internal_type() not in ('AutoField', 'BigAutoField'))
        new_temp_field.column = self._generate_temp_name(new_field.column)
        # Add it
        self.add_field(model, new_temp_field)
        # Explicit data type conversion
        # https://docs.oracle.com/en/database/oracle/oracle-database/18/sqlrf
        # /Data-Type-Comparison-Rules.html#GUID-D0C5A47E-6F93-4C2D-9E49-4F2B86B359DD
        new_value = self.quote_name(old_field.column)
        old_type = old_field.db_type(self.connection)
        if re.match('^N?CLOB', old_type):
            new_value = "TO_CHAR(%s)" % new_value
            old_type = 'VARCHAR2'
        if re.match('^N?VARCHAR2', old_type):
            new_internal_type = new_field.get_internal_type()
            if new_internal_type == 'DateField':
                new_value = "TO_DATE(%s, 'YYYY-MM-DD')" % new_value
            elif new_internal_type == 'DateTimeField':
                new_value = "TO_TIMESTAMP(%s, 'YYYY-MM-DD HH24:MI:SS.FF')" % new_value
            elif new_internal_type == 'TimeField':
                # TimeField are stored as TIMESTAMP with a 1900-01-01 date part.
                new_value = "TO_TIMESTAMP(CONCAT('1900-01-01 ', %s), 'YYYY-MM-DD HH24:MI:SS.FF')" % new_value
        # Transfer values across
        self.execute("UPDATE %s set %s=%s" % (
            self.quote_name(model._meta.db_table),
            self.quote_name(new_temp_field.column),
            new_value,
        ))
        # Drop the old field
        self.remove_field(model, old_field)
        # Rename and possibly make the new field NOT NULL
        super().alter_field(model, new_temp_field, new_field)

    def normalize_name(self, name):
        """
        Get the properly shortened and uppercased identifier as returned by
        quote_name() but without the quotes.
        """
        nn = self.quote_name(name)
        if nn[0] == '"' and nn[-1] == '"':
            nn = nn[1:-1]
        return nn

    def _generate_temp_name(self, for_name):
        """Generate temporary names for workarounds that need temp columns."""
        suffix = hex(hash(for_name)).upper()[1:]
        return self.normalize_name(for_name + "_" + suffix)

    def prepare_default(self, value):
        return self.quote_value(value)

    def _field_should_be_indexed(self, model, field):
        create_index = super()._field_should_be_indexed(model, field)
        db_type = field.db_type(self.connection)
        if db_type is not None and db_type.lower() in self.connection._limited_data_types:
            return False
        return create_index

    def _unique_should_be_added(self, old_field, new_field):
        return (
            super()._unique_should_be_added(old_field, new_field) and
            not self._field_became_primary_key(old_field, new_field)
        )

    def _is_identity_column(self, table_name, column_name):
        with self.connection.cursor() as cursor:
            cursor.execute("""
                SELECT
                    CASE WHEN identity_column = 'YES' THEN 1 ELSE 0 END
                FROM user_tab_cols
                WHERE table_name = %s AND
                      column_name = %s
            """, [self.normalize_name(table_name), self.normalize_name(column_name)])
            row = cursor.fetchone()
            return row[0] if row else False

    def _drop_identity(self, table_name, column_name):
        self.execute('ALTER TABLE %(table)s MODIFY %(column)s DROP IDENTITY' % {
            'table': self.quote_name(table_name),
            'column': self.quote_name(column_name),
        })



"""
SQL functions reference lists:
https://www.gaia-gis.it/gaia-sins/spatialite-sql-4.2.1.html
"""
from django.contrib.gis.db.backends.base.operations import (
    BaseSpatialOperations,
)
from django.contrib.gis.db.backends.spatialite.adapter import SpatiaLiteAdapter
from django.contrib.gis.db.backends.utils import SpatialOperator
from django.contrib.gis.db.models import aggregates
from django.contrib.gis.geos.geometry import GEOSGeometry, GEOSGeometryBase
from django.contrib.gis.geos.prototypes.io import wkb_r, wkt_r
from django.contrib.gis.measure import Distance
from django.core.exceptions import ImproperlyConfigured
from django.db.backends.sqlite3.operations import DatabaseOperations
from django.utils.functional import cached_property
from django.utils.version import get_version_tuple


class SpatialiteNullCheckOperator(SpatialOperator):
    def as_sql(self, connection, lookup, template_params, sql_params):
        sql, params = super().as_sql(connection, lookup, template_params, sql_params)
        return '%s > 0' % sql, params


class SpatiaLiteOperations(BaseSpatialOperations, DatabaseOperations):
    name = 'spatialite'
    spatialite = True

    Adapter = SpatiaLiteAdapter

    collect = 'Collect'
    extent = 'Extent'
    makeline = 'MakeLine'
    unionagg = 'GUnion'

    from_text = 'GeomFromText'

    gis_operators = {
        # Binary predicates
        'equals': SpatialiteNullCheckOperator(func='Equals'),
        'disjoint': SpatialiteNullCheckOperator(func='Disjoint'),
        'touches': SpatialiteNullCheckOperator(func='Touches'),
        'crosses': SpatialiteNullCheckOperator(func='Crosses'),
        'within': SpatialiteNullCheckOperator(func='Within'),
        'overlaps': SpatialiteNullCheckOperator(func='Overlaps'),
        'contains': SpatialiteNullCheckOperator(func='Contains'),
        'intersects': SpatialiteNullCheckOperator(func='Intersects'),
        'relate': SpatialiteNullCheckOperator(func='Relate'),
        'coveredby': SpatialiteNullCheckOperator(func='CoveredBy'),
        'covers': SpatialiteNullCheckOperator(func='Covers'),
        # Returns true if B's bounding box completely contains A's bounding box.
        'contained': SpatialOperator(func='MbrWithin'),
        # Returns true if A's bounding box completely contains B's bounding box.
        'bbcontains': SpatialOperator(func='MbrContains'),
        # Returns true if A's bounding box overlaps B's bounding box.
        'bboverlaps': SpatialOperator(func='MbrOverlaps'),
        # These are implemented here as synonyms for Equals
        'same_as': SpatialiteNullCheckOperator(func='Equals'),
        'exact': SpatialiteNullCheckOperator(func='Equals'),
        # Distance predicates
        'dwithin': SpatialOperator(func='PtDistWithin'),
    }

    disallowed_aggregates = (aggregates.Extent3D,)

    @cached_property
    def select(self):
        return 'CAST (AsEWKB(%s) AS BLOB)' if self.spatial_version >= (4, 3, 0) else 'AsText(%s)'

    function_names = {
        'ForcePolygonCW': 'ST_ForceLHR',
        'Length': 'ST_Length',
        'LineLocatePoint': 'ST_Line_Locate_Point',
        'NumPoints': 'ST_NPoints',
        'Reverse': 'ST_Reverse',
        'Scale': 'ScaleCoords',
        'Translate': 'ST_Translate',
        'Union': 'ST_Union',
    }

    @cached_property
    def unsupported_functions(self):
        unsupported = {'BoundingCircle', 'ForceRHR', 'MemSize'}
        if not self.lwgeom_version():
            unsupported |= {'Azimuth', 'GeoHash', 'IsValid', 'MakeValid'}
        return unsupported

    @cached_property
    def spatial_version(self):
        """Determine the version of the SpatiaLite library."""
        try:
            version = self.spatialite_version_tuple()[1:]
        except Exception as exc:
            raise ImproperlyConfigured(
                'Cannot determine the SpatiaLite version for the "%s" database. '
                'Was the SpatiaLite initialization SQL loaded on this database?' % (
                    self.connection.settings_dict['NAME'],
                )
            ) from exc
        if version < (4, 1, 0):
            raise ImproperlyConfigured('GeoDjango only supports SpatiaLite versions 4.1.0 and above.')
        return version

    def convert_extent(self, box):
        """
        Convert the polygon data received from SpatiaLite to min/max values.
        """
        if box is None:
            return None
        shell = GEOSGeometry(box).shell
        xmin, ymin = shell[0][:2]
        xmax, ymax = shell[2][:2]
        return (xmin, ymin, xmax, ymax)

    def geo_db_type(self, f):
        """
        Return None because geometry columns are added via the
        `AddGeometryColumn` stored procedure on SpatiaLite.
        """
        return None

    def get_distance(self, f, value, lookup_type):
        """
        Return the distance parameters for the given geometry field,
        lookup value, and lookup type.
        """
        if not value:
            return []
        value = value[0]
        if isinstance(value, Distance):
            if f.geodetic(self.connection):
                if lookup_type == 'dwithin':
                    raise ValueError(
                        'Only numeric values of degree units are allowed on '
                        'geographic DWithin queries.'
                    )
                dist_param = value.m
            else:
                dist_param = getattr(value, Distance.unit_attname(f.units_name(self.connection)))
        else:
            dist_param = value
        return [dist_param]

    def _get_spatialite_func(self, func):
        """
        Helper routine for calling SpatiaLite functions and returning
        their result.
        Any error occurring in this method should be handled by the caller.
        """
        cursor = self.connection._cursor()
        try:
            cursor.execute('SELECT %s' % func)
            row = cursor.fetchone()
        finally:
            cursor.close()
        return row[0]

    def geos_version(self):
        "Return the version of GEOS used by SpatiaLite as a string."
        return self._get_spatialite_func('geos_version()')

    def proj4_version(self):
        "Return the version of the PROJ.4 library used by SpatiaLite."
        return self._get_spatialite_func('proj4_version()')

    def lwgeom_version(self):
        """Return the version of LWGEOM library used by SpatiaLite."""
        return self._get_spatialite_func('lwgeom_version()')

    def spatialite_version(self):
        "Return the SpatiaLite library version as a string."
        return self._get_spatialite_func('spatialite_version()')

    def spatialite_version_tuple(self):
        """
        Return the SpatiaLite version as a tuple (version string, major,
        minor, subminor).
        """
        version = self.spatialite_version()
        return (version,) + get_version_tuple(version)

    def spatial_aggregate_name(self, agg_name):
        """
        Return the spatial aggregate SQL template and function for the
        given Aggregate instance.
        """
        agg_name = 'unionagg' if agg_name.lower() == 'union' else agg_name.lower()
        return getattr(self, agg_name)

    # Routines for getting the OGC-compliant models.
    def geometry_columns(self):
        from django.contrib.gis.db.backends.spatialite.models import SpatialiteGeometryColumns
        return SpatialiteGeometryColumns

    def spatial_ref_sys(self):
        from django.contrib.gis.db.backends.spatialite.models import SpatialiteSpatialRefSys
        return SpatialiteSpatialRefSys

    def get_geometry_converter(self, expression):
        geom_class = expression.output_field.geom_class
        if self.spatial_version >= (4, 3, 0):
            read = wkb_r().read

            def converter(value, expression, connection):
                return None if value is None else GEOSGeometryBase(read(value), geom_class)
        else:
            read = wkt_r().read
            srid = expression.output_field.srid
            if srid == -1:
                srid = None

            def converter(value, expression, connection):
                if value is not None:
                    geom = GEOSGeometryBase(read(value), geom_class)
                    if srid:
                        geom.srid = srid
                    return geom
        return converter
